/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package coalition;

import data.Attribute;
import data.AttributeType;
import data.Example;
import dbrsaaorc.params.Latex;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * Coalition class
 *
 * @author Mateusz Ledzianowski
 */
public class Coalition extends HashSet<Attribute> implements Comparable<Coalition> {

    public Coalition() {
        super();
    }

    /**
     * Basic constructor of the coalition
     *
     * @param ex1 example which attributes should be greater or equal to be in
     * the coalition
     * @param ex2 example which attributes should be lower to be in the
     * coalition
     */
    public Coalition(Example ex1, Example ex2) {
        super();
        calculateCoalition(ex1, ex2);
    }

    /**
     * Basic constructor of the coalition
     *
     * @param attributes set of coalition attributes
     */
    public Coalition(Set<Attribute> attributes) {
        super(attributes);
    }

    /**
     * Calculate coalition of attributes which are grater in ex1 than in ex2
     *
     * @param ex1 example which attributes should be greater or equal to be in
     * the coalition
     * @param ex2 example which attributes should be lower to be in the
     * coalition
     */
    private void calculateCoalition(Example ex1, Example ex2) {
        int n = ex1.getSize();
        if (n != ex2.getSize()) {
            throw new IllegalArgumentException("Both examples should have "
                    + "the same sizes!");
        }
        for (int i = 0; i < n; ++i) {
            Attribute attr = ex1.getAttribute(i);
            if (attr.getType() != AttributeType.OMIT && !attr.isDecision()) {
                Integer cmp_attr = attr.compareValues(ex1.getAttributeValue(i), ex2.getAttributeValue(i));
                if (cmp_attr != null && cmp_attr >= 0) {
                    this.add(attr);
                }
            }
        }
    }

    /**
     * @return hash code of the object
     */
    @Override
    public int hashCode() {
        return super.hashCode();
    }

    /**
     * @return true if objects are equal
     * @param obj object to compare
     */
    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final Coalition other = (Coalition) obj;
        for (Attribute attr : this) {
            if (!other.contains(attr)) {
                return false;
            }
        }
        return true;
    }

    /**
     * @param latex
     * @return string representation of the object
     */
    public String toString(Latex latex) {
        if (isEmpty()) {
            if (latex != Latex.NONE) {
                return "\\emptyset";
            } else {
                return "o";
            }
        }
        StringBuilder sb = new StringBuilder();
        if (latex != Latex.NONE) {
            sb.append("\\{");
        } else {
            sb.append("(");
        }
        List<Attribute> attributesList = new ArrayList<>(this);
        Collections.sort(attributesList);
        for (Attribute attr : attributesList) {
            sb.append(attr.getName(latex));
            if (latex != Latex.NONE) {
                sb.append(",\\allowbreak ");
            } else {
                sb.append(", ");
            }
        }
        if (!attributesList.isEmpty()) {
            if (latex != Latex.NONE) {
                sb.delete(sb.length() - 13, sb.length());
            } else {
                sb.delete(sb.length() - 2, sb.length());
            }
        }
        if (latex != Latex.NONE) {
            sb.append("\\}");
        } else {
            sb.append(")");
        }
        return sb.toString();
    }

    /**
     * @return string representation of the object
     */
    @Override
    public String toString() {
        return toString(Latex.NONE);
    }

    /**
     * Function compares two objects
     *
     * @param o other object
     * @return the value 0 if this is equal to @param o; a value less than 0 if
     * this is less than @param o; a value greater than 0 if this is greater
     * than @param o
     */
    @Override
    public int compareTo(Coalition o) {
        int cmp = Integer.valueOf(this.size()).compareTo(o.size());
        if (cmp == 0) {
            List<Attribute> myAttributes = new ArrayList<>(this);
            List<Attribute> otherAttributes = new ArrayList<>(o);
            Collections.sort(myAttributes);
            Collections.sort(otherAttributes);
            int n = myAttributes.size();
            for (int i = 0; i < n; ++i) {
                cmp = myAttributes.get(i).compareTo(otherAttributes.get(i));
                if (cmp != 0) {
                    return cmp;
                }
            }
        }
        return cmp;
    }

    /**
     * @param o other object
     * @return true if coalition is in coaliton o, false otherwise
     */
    public boolean in(Coalition o) {
        for (Attribute attr : this) {
            if (!o.contains(attr)) {
                return false;
            }
        }
        return true;
    }
}
