/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package coalition;

import data.Attribute;
import dbrsaaorc.params.Latex;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 *
 * @author Mateusz Ledzianowski
 */
public class SetOfCoalitions extends HashSet<Coalition> implements Comparable<SetOfCoalitions> {

    public SetOfCoalitions() {
        super();
    }

    public SetOfCoalitions(Set<Coalition> coalitions) {
        super(coalitions);
    }

    public SetOfCoalitions(SetOfCoalitions o) {
        super(new HashSet<Coalition>(o));
    }

    /**
     * Constructor of full sufficient coalitions
     *
     * @param attributes
     */
    public SetOfCoalitions(List<Attribute> attributes) {
        super();
        addAllPossibleAttributes(0, new Coalition(), attributes);
    }

    /**
     * Function calculates maximum coalitions of criteria.
     */
    protected final void calculateMaximumCoalitions() {
        List<Coalition> coalitionsList = new ArrayList<>(this);
        Collections.sort(coalitionsList);
        Collections.reverse(coalitionsList);
        this.clear();
        for (Coalition coalition : coalitionsList) {
            boolean isMaximum = true;
            for (Coalition coalitionInSet : this) {
                if (coalition.in(coalitionInSet)) {
                    isMaximum = false;
                    break;
                }
            }
            if (isMaximum) {
                this.add(coalition);
            }
        }
    }

    /**
     * Function calculates minimum coalitions of criteria.
     */
    protected final void calculateMinimumCoalitions() {
        List<Coalition> coalitionsList = new ArrayList<>(this);
        Collections.sort(coalitionsList);
        this.clear();
        for (Coalition coalition : coalitionsList) {
            boolean isMinimum = true;
            for (Coalition otherCoalition : this) {
                if (otherCoalition.in(coalition)) {
                    isMinimum = false;
                    break;
                }
            }
            if (isMinimum) {
                this.add(coalition);
            }
        }
    }

    /**
     * Function remove covered coalitions from the set
     *
     * @param coverCoalitions
     * @param allAttributes
     */
    protected final void removeCoveredCoalitions(
            SetOfCoalitions coverCoalitions,
            List<Attribute> allAttributes) {
        SetOfCoalitions setOfCoalitionsClone = (SetOfCoalitions) this.clone();
        this.clear();
        for (Coalition coalition : setOfCoalitionsClone) {
            boolean covered = false;
            for (Coalition mic : coverCoalitions) {
                if (coalition.in(mic)) {
                    covered = true;
                    break;
                }
            }
            if (!covered) {
                this.add(coalition);
            }
        }
        if (this.isEmpty() && !coverCoalitions.isEmpty()) {
            addMinimumNotCoveredCoalitions(coverCoalitions, allAttributes);
        }
    }

    private void addMinimumNotCoveredCoalitions(SetOfCoalitions coverCoalitions,
            List<Attribute> allAttributes) {
        for (Coalition coverCoalition : coverCoalitions) {
            List<Attribute> notCoveredAttributes = new ArrayList<>();
            for (Attribute attr : allAttributes) {
                if (!coverCoalition.contains(attr)) {
                    notCoveredAttributes.add(attr);
                }
            }
            if (this.isEmpty()) {
                for (Attribute attr : notCoveredAttributes) {
                    Coalition newCoalition = new Coalition();
                    newCoalition.add(attr);
                    this.add(newCoalition);
                }
            } else if (notCoveredAttributes.size() == 1) {
                for (Coalition coalition : this) {
                    coalition.add(notCoveredAttributes.get(0));
                }
            } else {
                SetOfCoalitions oldCoalitions = (SetOfCoalitions) this.clone();
                this.clear();
                for (Coalition coalition : oldCoalitions) {
                    for (Attribute attr : notCoveredAttributes) {
                        Coalition newCoalition = new Coalition(coalition);
                        newCoalition.add(attr);
                        this.add(newCoalition);
                    }
                }
            }
        }
    }

    private void addAllPossibleAttributes(int i, Coalition coalition, List<Attribute> attributes) {
        if (i < attributes.size()) {
            Coalition withCoalition = (Coalition) coalition.clone();
            withCoalition.add(attributes.get(i));
            addAllPossibleAttributes(i + 1, withCoalition, attributes);
            Coalition withoutCoalition = (Coalition) coalition.clone();
            addAllPossibleAttributes(i + 1, withoutCoalition, attributes);
        }
        super.add(coalition);
    }

    /**
     * @param latex
     * @return string representation of the object
     */
    public String toString(Latex latex) {
        StringBuilder sb = new StringBuilder();
        if (latex == Latex.NONE) {
            sb.append("{");
        }
        List<Coalition> coalitionsList = new ArrayList<>(this);
        Collections.sort(coalitionsList);
        for (Coalition coalition : coalitionsList) {
            sb.append(coalition.toString(latex));
            if (latex != Latex.NONE) {
                sb.append(",\\allowbreak ");
            } else {
                sb.append(", ");
            }
        }
        if (!coalitionsList.isEmpty()) {
            if (latex != Latex.NONE) {
                sb.delete(sb.length() - 13, sb.length());
            } else {
                sb.delete(sb.length() - 2, sb.length());
            }
        }
        if (latex == Latex.NONE) {
            sb.append("}");
        }
        return sb.toString();
    }

    /**
     * @return string representation of the object
     */
    @Override
    public String toString() {
        return toString(Latex.NONE);
    }

    @Override
    public int compareTo(SetOfCoalitions o) {
        int cmp = Integer.compare(this.size(), o.size());
        if (cmp == 0) {
            List<Coalition> coalitionList = new ArrayList<>(this);
            Collections.sort(coalitionList);
            List<Coalition> otherCoalitionList = new ArrayList<>(o);
            Collections.sort(otherCoalitionList);
            for (int i = 0; i < this.size(); ++i) {
                cmp = coalitionList.get(i).compareTo(otherCoalitionList.get(i));
                if (cmp != 0) {
                    return cmp;
                }
            }
        }
        return cmp;
    }
}
