/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package data;

import dbrsaaorc.params.Latex;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import utils.Lang;
import utils.LatexHelper;

/**
 * Attribute class
 *
 * @author Mateusz Ledzianowski
 */
public class Attribute implements Comparable<Attribute> {

    /**
     * A name of the attribute
     */
    private String name;
    /**
     * A type of the attribute
     */
    private AttributeType type;
    /**
     * True if the attribute is a decision
     */
    private boolean decision;
    /**
     * Possible values of attribute - the key is string value and value is
     * ordinal number
     */
    private List<String> values;
    /**
     * Preference of attribute
     */
    private AttributePreference preference;

    /**
     * Basic constructor of the attribute
     *
     * @param line containing name and values of attribute in isf format
     */
    public Attribute(String line) {
        line = line.replaceAll(",", " ");
        line = line.replaceAll("\t", " ");
        line = line.replaceAll(" +", " ");
        if (line.startsWith("+ ")) {
            line = line.substring(2);
        }
        String[] splitLine = line.split("[, \t]");
        name = splitLine[0].substring(0,
                splitLine[0].indexOf(":"));
        type = AttributeType.getAttributeType(splitLine[1]);
        decision = false;
        if (type == AttributeType.ORDEREDLIST
                || type == AttributeType.UNORDEREDLIST) {
            int n = splitLine.length;
            splitLine[1] = splitLine[1].substring(1);
            splitLine[n - 1] = splitLine[n - 1]
                    .substring(0, splitLine[n - 1].length() - 1);
            values = Arrays.asList(splitLine).subList(1, n);
        } else {
            values = new ArrayList<>();
        }
        preference = AttributePreference.GAIN;
    }

    /**
     * Basic constructor of the attribute
     *
     * @param name of the attribute
     * @param type of the attribute
     */
    public Attribute(String name, AttributeType type) {
        this.name = name;
        this.type = type;
        this.decision = false;
        values = new ArrayList<>();
    }

    /**
     * Basic constructor of the attribute
     *
     * @param name of the attribute
     * @param type of the attribute
     * @param decision is True if the attribute is a decision
     */
    public Attribute(String name, AttributeType type, boolean decision) {
        this.name = name;
        this.type = type;
        this.decision = decision;
        this.type = AttributeType.OMIT;
        values = new ArrayList<>();
    }

    /**
     * Basic constructor of the attribute
     *
     * @param name of the attribute
     * @param type of the attribute
     * @param values of the attribute (strings)
     */
    public Attribute(String name, AttributeType type, List<String> values) {
        this.name = name;
        this.type = type;
        this.decision = false;
        this.values = values;
    }

    /**
     * Basic constructor of the attribute
     *
     * @param name of the attribute
     * @param type of the attribute
     * @param decision is True if the attribute is a decision
     * @param values of the attribute (strings)
     */
    public Attribute(String name, AttributeType type, boolean decision,
            List<String> values) {
        this.name = name;
        this.type = type;
        this.decision = decision;
        this.values = values;
    }

    /**
     * Filter set of attributes to only this which are not omit neither
     * decision.
     *
     * @param attributes set of attributes to filter
     *
     * @return set of attributes which are not omit neither decision.
     */
    public static Set<Attribute> getSignificantAttributes(Set<Attribute> attributes) {
        Set<Attribute> significantAttributes = new HashSet<>();
        for (Attribute attribute : attributes) {
            if (!attribute.isDecision() && attribute.getType() != AttributeType.OMIT) {
                significantAttributes.add(attribute);
            }
        }
        return significantAttributes;
    }

    /**
     * Filter list of attributes to only this which are not omit neither
     * decision.
     *
     * @param attributes list of attributes to filter
     *
     * @return list of attributes which are not omit neither decision.
     */
    public static List<Attribute> getSignificantAttributes(List<Attribute> attributes) {
        List<Attribute> significantAttributes = new ArrayList<>();
        for (Attribute attribute : attributes) {
            if (!attribute.isDecision()
                    && attribute.getType() != AttributeType.OMIT
                    && attribute.getType() != AttributeType.UNORDEREDLIST) {
                significantAttributes.add(attribute);
            }
        }
        return significantAttributes;
    }

    /**
     * Compare two values of the attribute
     *
     * @param value1 to compare (string)
     * @param value2 to compare (string)
     * @return the value 0 if value1 is equal to value2; a value less than 0 if
     * value1 is less than value2; a value greater than 0 if value1 is greater
     * than value2; null if values are uncomparable
     */
    public Integer compareValues(String value1, String value2) {
        Integer cmp;
        switch (type) {
            case CONTINUOUS:
                cmp = Double.valueOf(value1).compareTo(Double.valueOf(value2));
                break;
            case NUMBERCODED:
                cmp = Integer.valueOf(value1).compareTo(Integer.valueOf(value2));
                break;
            case ORDEREDLIST:
                cmp = getValuePosition(value1).compareTo(getValuePosition(value2));
                break;
            case UNORDEREDLIST:
                if (decision) {
                    cmp = getValuePosition(value1).compareTo(getValuePosition(value2));
                    break;
                }
            default:
                return null;
        }
        if (getPreference() == AttributePreference.GAIN) {
            return cmp;
        } else {
            return -cmp;
        }
    }

    /**
     * @return string of attribute's values
     */
    private String valuesString(Latex latex) {
        StringBuilder sb = new StringBuilder();
        boolean first = true;
        for (String str : values) {
            if (first) {
                first = false;
            } else {
                if (latex != Latex.NONE) {
                    sb.append(",\\allowbreak ");
                } else {
                    sb.append(", ");
                }
            }
            sb.append(str);
        }
        return sb.toString();
    }

    /**
     * @return string of attribute's type
     */
    private String valuesType(Latex latex) {
        switch (type) {
            case CONTINUOUS:
                return "(continuous)";
            case NUMBERCODED:
                return "(numbercoded)";
            case OMIT:
                return "(omit)";
            case ORDEREDLIST:
                return "<" + valuesString(latex) + ">";
            default:
                return "[" + valuesString(latex) + "]";
        }
    }

    /**
     * @param value which position is under consideration
     * @return position of the value
     */
    public Integer getValuePosition(String value) {
        switch (type) {
            case CONTINUOUS:
                Double pos = Double.valueOf(value) * 1000;
                return pos.intValue();
            case NUMBERCODED:
                return Integer.valueOf(value);
            case OMIT:
                return null;
            default:
                return values.indexOf(value);
        }
    }

    @Override
    public int hashCode() {
        int hashFirst = name != null ? name.hashCode() : 0;
        int hashSecond = type != null ? type.hashCode() : 0;

        return (hashFirst + hashSecond) * hashSecond + hashFirst;
    }

    @Override
    public boolean equals(Object other) {
        if (other instanceof Attribute) {
            Attribute otherAttribute = (Attribute) other;
            return (((this.name == null
                    ? otherAttribute.name == null
                    : this.name.equals(otherAttribute.name))
                    || (this.name != null && otherAttribute.name != null
                    && this.name.equals(otherAttribute.name)))
                    && (this.type == otherAttribute.type
                    || (this.type != null && otherAttribute.type != null
                    && this.type.equals(otherAttribute.type))));
        }

        return false;
    }

    public String toString(Latex latex) {
        if (latex != Latex.NONE) {
            StringBuilder sb = new StringBuilder();
            if (isDecision()) {
                sb.append(LatexHelper.equation("\\bm{" + LatexHelper.nameToString(name) + "}", latex));
            } else {
                sb.append(LatexHelper.equation(LatexHelper.nameToString(name), latex));
            }
            sb.append(" & ");
            if (isDecision()) {
                sb.append("\\bm{");
                sb.append(LatexHelper.equation(valuesType(latex), latex));
                sb.append("}");
            } else {
                sb.append(LatexHelper.equation(valuesType(latex), latex));
            }
            sb.append(" & ");
            if (isDecision()) {
                sb.append("\\bm{");
                if (getPreference() == AttributePreference.GAIN) {
                    sb.append(LatexHelper.equation(Lang.gain(latex), latex));
                } else {
                    sb.append(LatexHelper.equation(Lang.cost(latex), latex));
                }
                sb.append("}");
            } else {
                if (getPreference() == AttributePreference.GAIN) {
                    sb.append(LatexHelper.equation(Lang.gain(latex), latex));
                } else {
                    sb.append(LatexHelper.equation(Lang.cost(latex), latex));
                }
            }
            if (latex != Latex.NONE) {
                sb.append(" \\\\");
            }
            return sb.toString();
        } else {
            return name + ": " + valuesType(latex);
        }
    }

    @Override
    public String toString() {
        return toString(Latex.NONE);
    }

    /**
     * Function compares two objects
     *
     * @param o other object
     * @return the value 0 if this is equal to @param o; a value less than 0 if
     * this is less than @param o; a value greater than 0 if this is greater
     * than @param o
     */
    @Override
    public int compareTo(Attribute o) {
        int cmp = this.name.compareTo(o.name);
        if (cmp == 0) {
            return this.type.compareTo(o.type);
        } else {
            return cmp;
        }
    }

    public String getNameWithPreference(Latex latex) {
        if (latex != Latex.NONE) {
            StringBuilder sb = new StringBuilder(LatexHelper.nameToString(name) + " ");
            if (preference == AttributePreference.GAIN) {
                sb.append("\\uparrow");
            }
            if (preference == AttributePreference.COST) {
                sb.append("\\downarrow");
            }
            return sb.toString();
        } else {
            StringBuilder sb = new StringBuilder(name + " ");
            if (preference == AttributePreference.GAIN) {
                sb.append("gain");
            }
            if (preference == AttributePreference.COST) {
                sb.append("cost");
            }
            return sb.toString();
        }
    }

    public String getName(Latex latex) {
        if (latex != Latex.NONE) {
            return LatexHelper.nameToString(name);
        } else {
            return name;
        }
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

    public boolean isDecision() {
        return decision;
    }

    public void setDecision(boolean decision) {
        this.decision = decision;
    }

    public AttributeType getType() {
        return type;
    }

    public void setType(AttributeType type) {
        this.type = type;
    }

    public List<String> getValues() {
        return values;
    }

    public void setValues(List<String> values) {
        this.values = values;
    }

    public AttributePreference getPreference() {
        return preference;
    }

    public void setPreference(AttributePreference preference) {
        this.preference = preference;
    }

    public boolean isCriterion() {
        if (isDecision()) {
            return false;
        }
        if (getType() == AttributeType.OMIT) {
            return false;
        }
        if (getType() == AttributeType.UNORDEREDLIST) {
            return false;
        }
        return true;
    }
}
