/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package data;

import dbrsaaorc.params.Latex;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;
import utils.Lang;
import utils.LatexHelper;

/**
 * ISF file parser class
 *
 * @author Mateusz Ledzianowski
 */
public class ISF {

    /**
     * Name of dataset
     */
    private String name;
    /**
     * List of attributes
     */
    private final List<Attribute> attributes;
    /**
     * List of examples
     */
    private final List<Example> examples;

    /**
     * Basic constructor of ISF file parser
     */
    public ISF() {
        attributes = new ArrayList<>();
        examples = new ArrayList<>();
    }

    /**
     * Basic constructor of ISF file parser which load input from a given file
     *
     * @param path to a file with data
     * @throws IOException if a file does not exists or could not be open
     */
    public ISF(String path) throws IOException {
        File f = new File(path);
        name = f.getName().substring(0, f.getName().lastIndexOf("."));
        attributes = new ArrayList<>();
        examples = new ArrayList<>();

        ISFState state = ISFState.BEGIN;
        Set<String> decisions = new HashSet<>();

        File file = new File(path);
        try (FileReader fileReader = new FileReader(file)) {
            BufferedReader bufferedReader = new BufferedReader(fileReader);
            String line;
            while ((line = getSignificantLine(bufferedReader)) != null) {
                StringBuilder sb = new StringBuilder(line);
                while (!isEndedLine(line)
                        && (line = getSignificantLine(bufferedReader)) != null) {
                    while (line != null && line.isEmpty()) {
                        line = getSignificantLine(bufferedReader);
                    }
                    sb.append(line);
                }
                line = getClearedLine(sb.toString());

                if (!line.isEmpty()) {
                    switch (state) {
                        case BEGIN:
                            if (line.equals("**ATTRIBUTES")) {
                                state = ISFState.ATTRIBUTES;
                            } else if (line.equals("**END")) {
                                state = ISFState.END;
                            }
                            break;
                        case ATTRIBUTES:
                            switch (line) {
                                case "**EXAMPLES":
                                    updateDecisionAttributes(decisions);
                                    state = ISFState.EXAMPLES;
                                    break;
                                case "**PREFERENCES":
                                    updateDecisionAttributes(decisions);
                                    state = ISFState.PREFERENCES;
                                    break;
                                case "**END":
                                    updateDecisionAttributes(decisions);
                                    state = ISFState.END;
                                    break;
                                default:
                                    if (line.startsWith("decision:")) {
                                        decisions.add(line.split("[, \t]")[1]);
                                    } else {
                                        attributes.add(new Attribute(line));
                                    }
                                    break;
                            }
                            break;
                        case PREFERENCES:
                            if (line.equals("**EXAMPLES")) {
                                state = ISFState.EXAMPLES;
                                break;
                            } else if (line.equals("**END")) {
                                state = ISFState.END;
                                break;
                            }
                            updateExamplePreference(line);
                            break;
                        case EXAMPLES:
                            if (line.equals("**END")) {
                                state = ISFState.END;
                            } else {
                                examples.add(new Example(attributes, line));
                            }
                            break;
                    }
                }
            }
            if (examples.size() > 0) {
                Attribute decisionAttribute = examples.get(0)
                        .getDecisionAttribute();
                for (String decisionValue
                        : decisionAttribute.getValues()) {
                }
            }
        } catch (IOException ex) {
            Logger.getLogger(ISF.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    /**
     * @param latex
     * @param labelSuff suffix of tables label (only if latex is not NONE)
     * @return string representation of the object
     */
    public String toString(Latex latex, String labelSuff) {
        StringBuilder sb = new StringBuilder();
        if (latex != Latex.NONE) {
            sb.append("\\begin{table}[H]\n"
                    + "\\centering\n"
                    + "\\begin{tabularx}{\\textwidth}{@{}XXX@{}}\n"
                    + "\\toprule\n"
                    + Lang.attribute(latex) + " & " + Lang.description(latex) + " & " + Lang.type(latex) + " \\\\ \\midrule\n");
        } else {
            sb.append("**ATTRIBUTES\n");
            getAttributes().stream().filter(x -> x.isDecision())
                    .forEach(x -> {
                        sb.append("decision: ");
                        sb.append(x.getName(latex));
                        sb.append("\n");
                    });
        }
        for (Attribute attr : getAttributes()) {
            sb.append(attr.toString(latex));
            sb.append("\n");
        }
        if (latex != Latex.NONE) {
            sb.setLength(sb.length() - 1);
            sb.append(" \\bottomrule\n"
                    + "\\end{tabularx}\n"
                    + "\\caption{" + Lang.attributesCaption(latex)
                    + " \\textit{" + name + "}.}\n"
                    + "\\label{attr-" + labelSuff + "}\n"
                    + "\\end{table}\n\n");
        } else {
            if (getAttributes().stream()
                    .anyMatch(x
                            -> x.getPreference() == AttributePreference.COST)) {
                sb.append("**PREFERENCES\n");
                for (Attribute attr : getAttributes()) {
                    sb.append(LatexHelper.equation(attr.getName(latex), latex));
                    sb.append(": ");
                    sb.append(attr.getPreference().toString());
                    sb.append("\n");
                }
            }
        }
        if (latex != Latex.NONE) {
            sb.append("\\begin{table}[H]\n"
                    + "\\centering\n"
                    + "\\begin{tabularx}{\\textwidth}{@{}"
                    + LatexHelper.repeat("X", getAttributes().size())
                    + "@{}}\n"
                    + "\\toprule\n");
            for (Attribute attr : getAttributes()) {
                if (attr.isDecision()) {
                    sb.append(LatexHelper.equation("\\bm{"
                            + attr.getNameWithPreference(latex) + "}", latex));
                } else {
                    sb.append(LatexHelper.equation(
                            attr.getNameWithPreference(latex), latex));
                }
                sb.append(" & ");
            }
            sb.setLength(sb.length() - 2);
            sb.append("\\\\ \\midrule\n");
        } else {
            sb.append("**EXAMPLES\n");
        }
        for (Example example : getExamples()) {
            sb.append(example.toString(latex));
            sb.append("\n");
        }
        if (latex != Latex.NONE) {
            sb.setLength(sb.length() - 1);
            sb.append(" \\bottomrule\n"
                    + "\\end{tabularx}\n"
                    + "\\caption{" + Lang.examplesCaption(latex)
                    + " \\textit{" + name + "}.}\n"
                    + "\\label{ex-" + labelSuff + "}\n"
                    + "\\end{table}\n");
        } else {
            sb.append("**END\n");
        }
        return sb.toString();
    }

    /**
     * @return string representation of the object
     */
    @Override
    public String toString() {
        return toString(Latex.NONE, "");
    }

    /**
     * @param bufferedReader from which lines are loaded
     * @return next not empty line from bufferedReader without comments
     */
    protected static String getSignificantLine(BufferedReader bufferedReader) {
        try {
            String line = bufferedReader.readLine();
            if (line == null) {
                return null;
            }
            line = line.split("[|#!]", 2)[0];
            while (line != null && line.isEmpty()) {
                line = bufferedReader.readLine();
                if (line == null) {
                    return null;
                }
                line = line.split("[|#!]", 2)[0];
            }
            return line;
        } catch (IOException ex) {
            Logger.getLogger(ISF.class.getName()).log(Level.SEVERE, null, ex);
            return null;
        }
    }

    /**
     * @param line to check
     * @return true if line is ended (in isf format backslash (/) at the end of
     * line means that it will be continued after a sign of a new line)
     */
    protected static boolean isEndedLine(String line) {
        for (int i = line.length() - 1; i >= 0; --i) {
            if (line.charAt(i) == '\\') {
                return false;
            } else if (line.charAt(i) != ' ') {
                return true;
            }
        }
        return true;
    }

    /**
     * @param line without comments and merged if was broken
     * @return cleared line to format with only one separator sign between
     * fields and no other characters.
     */
    protected static String getClearedLine(String line) {
        return line.replace("\\", "")
                .replaceAll("\r\n", "")
                .replaceAll("\n", "")
                .replaceAll("[ \t]+", " ")
                .replaceAll("^ ", "")
                .replaceAll(" $", "")
                .replaceAll("[.]$", "")
                .replaceAll(", ", ",")
                .replaceAll(" ,", ",")
                .replaceAll(" ", ",");
    }

    /**
     * Function sets given attributes as decisions
     *
     * @param decisions - list of attributes names to set as decision
     */
    private void updateDecisionAttributes(Set<String> decisions) {
        for (String str : decisions) {
            Attribute attr = getAttributes().stream()
                    .filter(x -> x.getName().equals(str))
                    .findFirst().orElse(null);
            if (attr != null) {
                attr.setDecision(true);
            }
        }
    }

    private void updateExamplePreference(String line) {
        String[] lineTab = line.split(":,");
        Attribute attr = getAttributeByName(lineTab[0]);
        if (attr != null) {
            attr.setPreference(AttributePreference.getAttributePreference(lineTab[1]));
        }
    }

    /**
     * @return the list of attributes
     */
    public List<Attribute> getAttributes() {
        return attributes;
    }

    /**
     * @return the list of attributes
     */
    public List<Attribute> getCriteria() {
        return attributes.stream().filter(x -> x.isCriterion())
                .collect(Collectors.toList());
    }

    /**
     * @return the list of examples
     */
    public List<Example> getExamples() {
        return examples;
    }

    /**
     * @return decision attribute
     */
    public Attribute getDecisionAttribute() {
        return getAttributes().stream()
                .filter(x -> x.isDecision())
                .findFirst().orElse(null);
    }

    private Attribute getAttributeByName(String attrName) {
        return attributes.stream().filter(x -> x.getName().equals(attrName)).findFirst().orElse(null);
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }
}
