/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package domlem;

import coalition.Coalition;
import coalition.SetOfCoalitions;
import data.Example;
import core.SetOfExamples;
import data.Attribute;
import dbrsaaorc.params.Latex;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import utils.LatexHelper;

/**
 *
 * @author Mateusz Ledzianowski
 */
public class Rule {

    private static long idSeq = 0;

    public static Rule getBestRule(Map<Rule, RuleScore> rulesValues) {
        Rule resultRule = rulesValues.keySet().stream().findFirst().orElse(null);
        float bestCertainty = rulesValues.get(resultRule).getCertainty();
        int bestSupport = rulesValues.get(resultRule).getSupport();
        for (Rule rule : rulesValues.keySet()) {
            if (rulesValues.get(rule).getCertainty() > bestCertainty) {
                resultRule = rule;
                bestCertainty = rulesValues.get(rule).getCertainty();
                bestSupport = rulesValues.get(rule).getSupport();
            } else if (rulesValues.get(rule).getCertainty() == bestCertainty && rulesValues.get(rule).getSupport() > bestSupport) {
                resultRule = rule;
                bestCertainty = rulesValues.get(rule).getCertainty();
                bestSupport = rulesValues.get(rule).getSupport();
            }
        }

        return resultRule;
    }

    private final long id;
    private final List<Condition> conditions;
    private final Decision decision;
    public final boolean greaterOrEqual;
    private RuleScore ruleScore;
    private RuleUnionScore ruleUnionScore;

    public Rule(List<Condition> conditions, Decision decision) {
        this.id = idSeq++;
        this.conditions = conditions;
        this.decision = decision;
        this.greaterOrEqual = decision.isGreaterOrEqual();
    }

    public Rule(Rule rule) {
        this.id = idSeq++;
        this.conditions = new ArrayList<>(rule.conditions);
        this.decision = rule.decision;
        this.greaterOrEqual = rule.greaterOrEqual;
    }

    public Rule addBestCondition(SetOfCoalitions minimumSufficientCoalitions, SetOfExamples examplesToCover, SetOfExamples examplesNotCovered) {
        Map<Attribute, Set<Rule>> rules = new HashMap<>(); // All possible conditions for all attributes
        Map<Rule, RuleScore> rulesValues = new HashMap<>(); // Values of all conditions
        Set<Attribute> attributes = new HashSet<>();
        for (Coalition coalition : minimumSufficientCoalitions) {
            boolean candidate = true;
            for (Condition cond : conditions) {
                if (!coalition.contains(cond.getAttribute())) {
                    candidate = false;
                }
            }
            if (candidate) {
                for (Attribute attr : coalition) {
                    attributes.add(attr);
                }
            }
        }
        for (Attribute attr : attributes) {
            boolean hasAttribute = false;
            for (Condition cond : this.conditions) {
                if (cond.getAttribute().equals(attr)) {
                    hasAttribute = true;
                    break;
                }
            }
            if (!hasAttribute) {
                Set<Rule> rulesSet = new HashSet<>();
                for (Example ex : examplesToCover) {
                    Rule rule = new Rule(this);
                    rule.conditions.add(new Condition(attr, ex.getAttributeValue(attr), greaterOrEqual));
                    if (!rulesSet.contains(rule)) {
                        rulesSet.add(rule);
                        rulesValues.put(rule, new RuleUnionScore(rule, examplesToCover, examplesNotCovered));
                    }
                }
                rules.put(attr, rulesSet);
            }
        }

        if (rulesValues.isEmpty()) {
            return this;
        }
        return getBestRule(rulesValues);
    }

    public Rule getSimplifiedRule(SetOfCoalitions minimumSufficientCoalitions, SetOfExamples classesUnion, SetOfExamples examples) {
        Rule resultRule = this;
        int simplifiedIndex = 0;
        while (simplifiedIndex < resultRule.conditions.size()) {
            Rule newRule = new Rule(resultRule);
            newRule.conditions.remove(simplifiedIndex);
            if (newRule.isCompleted(classesUnion, examples) && newRule.isSufficient(minimumSufficientCoalitions)) {
                resultRule = newRule;
                --simplifiedIndex;
            }
            ++simplifiedIndex;
        }
        return resultRule;
    }

    public boolean covers(Example ex) {
        for (Condition cond : getConditions()) {
            if (!cond.isFulfilled(ex)) {
                return false;
            }
        }
        return true;
    }

    /**
     * @param examplesToCover
     * @param examplesNotCovered
     * @return true if rule certainty is 100%.
     */
    public boolean isCompleted(SetOfExamples examplesToCover, SetOfExamples examplesNotCovered) {
        return getRuleUnionScore(examplesToCover, examplesNotCovered).getCertainty() >= 0.999999;
    }

    boolean isSufficient(SetOfCoalitions minimumSufficientCoalitions) {
        Coalition ruleCoalition = new Coalition();
        for (Condition cond : this.conditions) {
            ruleCoalition.add(cond.getAttribute());
        }
        for (Coalition coalition : minimumSufficientCoalitions) {
            if (coalition.in(ruleCoalition)) {
                return true;
            }
        }
        return false;
    }

    /**
     * @param ex example to check
     * @return true if all conditions are fullfilled, false otherwise
     */
    public boolean isFulfilled(Example ex) {
        for (Condition cond : conditions) {
            if (!cond.isFulfilled(ex)) {
                return false;
            }
        }
        return true;
    }

    @Override
    public int hashCode() {
        int hash = 3;
        hash = 17 * hash + Objects.hashCode(this.conditions);
        hash = 17 * hash + Objects.hashCode(this.decision);
        return hash;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final Rule other = (Rule) obj;
        for (Condition cond : this.conditions) {
            if (!other.conditions.contains(cond)) {
                return false;
            }
        }
        return Objects.equals(this.decision, other.decision);
    }

    public String toString(Latex latex, boolean sortedConditions) {
        if (latex != Latex.NONE) {
            StringBuilder sb = new StringBuilder(
                    LatexHelper.equation(Long.toString(id), latex));
            sb.append(" & ");
            sb.append(LatexHelper.equation(conditionsString(latex, sortedConditions), latex));
            sb.append(" & ");
            sb.append(LatexHelper.equation(decision.toShortString(latex), latex));
            sb.append(" & ");
            if (!getRuleScore().getExamplesInCondition().isEmpty()) {
                sb.append(LatexHelper.equation(
                        getRuleScore().getExamplesInCondition().toString(latex), latex));
            }
            return sb.toString();
        } else {
            return "if " + conditionsString(latex, sortedConditions)
                    + " then "
                    + LatexHelper.equation(decision.toShortString(latex), latex)
                    + " (" + id + ")";
        }
    }

    @Override
    public String toString() {
        return toString(Latex.NONE, true);
    }

    private String conditionsString(Latex latex, boolean sorted) {
        if (conditions.isEmpty()) {
            if (latex != Latex.NONE) {
                return "\\emptyset";
            } else {
                return "o";
            }
        }
        StringBuilder sb = new StringBuilder();
        List<Condition> tmpConditions = new ArrayList<>(conditions);
        if (sorted) {
            Collections.sort(tmpConditions);
        }
        sb.append(tmpConditions.get(0).toShortString(latex));
        for (int i = 1; i < tmpConditions.size(); ++i) {
            if (latex != Latex.NONE) {
                sb.append(" \\wedge ");
            } else {
                sb.append(" & ");
            }
            sb.append(tmpConditions.get(i).toShortString(latex));
        }
        return sb.toString();
    }

    public List<Condition> getConditions() {
        return conditions;
    }

    public Decision getDecision() {
        return decision;
    }

    public boolean isGreaterOrEqual() {
        return greaterOrEqual;
    }

    public long getId() {
        return id;
    }

    public RuleScore getRuleScore() {
        return ruleScore;
    }

    public RuleScore getRuleScore(SetOfExamples examples) {
        if (ruleScore == null) {
            ruleScore = new RuleScore(this, examples);
        }
        return ruleScore;
    }

    public void setRuleScore(SetOfExamples examples) {
        this.ruleScore = new RuleScore(this, examples);
    }

    public RuleUnionScore getRuleUnionScore(SetOfExamples examplesToCover, SetOfExamples examples) {
        if (ruleUnionScore == null) {
            ruleUnionScore = new RuleUnionScore(this, examplesToCover, examples);
        }
        return ruleUnionScore;
    }

    public void setRuleUnionScore(SetOfExamples examplesToCover, SetOfExamples examples) {
        this.ruleUnionScore = new RuleUnionScore(this, examplesToCover, examples);
    }
}
